@@ -11,15 +11,23 @@ import (
 	"testing"
 	"time"
 
-	"github.com/openshift/service-ca-operator/pkg/controller/api"
-	"github.com/openshift/service-ca-operator/pkg/operator/operatorclient"
-
 	v1 "k8s.io/api/core/v1"
 	"k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
+	kruntime "k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/apimachinery/pkg/util/wait"
 	"k8s.io/client-go/kubernetes"
+	"k8s.io/client-go/rest"
 	"k8s.io/client-go/tools/clientcmd"
+
+	operatorv1client "github.com/openshift/client-go/operator/clientset/versioned"
+	"github.com/openshift/library-go/pkg/crypto"
+	"github.com/openshift/library-go/pkg/operator/events"
+	"github.com/openshift/library-go/pkg/operator/resource/resourceapply"
+	"github.com/openshift/service-ca-operator/pkg/controller/api"
+	"github.com/openshift/service-ca-operator/pkg/operator"
+	"github.com/openshift/service-ca-operator/pkg/operator/operatorclient"
+	"github.com/openshift/service-ca-operator/test/util"
 )
 
 const (
@@ -30,22 +38,43 @@ const (
 	configMapInjectorPodPrefix   = api.ConfigMapInjectorDeploymentName
 	caControllerPodPrefix        = api.SignerControllerDeploymentName
 	signingKeySecretName         = api.SignerControllerSecretName
+
+	pollInterval = time.Second
+	pollTimeout  = 30 * time.Second
+
+	// Rotation of all certs and bundles is expected to take a considerable amount of time
+	// due to the operator having to restart each controller and then each controller having
+	// to acquire the leader election lease and update all targeted resources.
+	rotationTimeout = 3 * time.Minute
 )
 
-func hasPodWithPrefixName(client *kubernetes.Clientset, name, namespace string) bool {
-	if client == nil || len(name) == 0 || len(namespace) == 0 {
-		return false
+// checkComponents verifies that the components of the operator are running.
+func checkComponents(t *testing.T, client *kubernetes.Clientset) {
+	componentConfigs := []struct {
+		namespace string
+		podPrefix string
+	}{
+		{serviceCAOperatorNamespace, serviceCAOperatorPodPrefix},
+		{serviceCAControllerNamespace, apiInjectorPodPrefix},
+		{serviceCAControllerNamespace, configMapInjectorPodPrefix},
+		{serviceCAControllerNamespace, caControllerPodPrefix},
 	}
-	pods, err := client.CoreV1().Pods(namespace).List(metav1.ListOptions{})
-	if err != nil {
-		return false
-	}
-	for _, pod := range pods.Items {
-		if strings.HasPrefix(pod.GetName(), name) {
-			return true
+	for _, cfg := range componentConfigs {
+		pods, err := client.CoreV1().Pods(cfg.namespace).List(metav1.ListOptions{})
+		if err != nil {
+			t.Fatalf("Failed to list pods in namespace %q: %v", cfg.namespace, err)
+		}
+		podFound := false
+		for _, pod := range pods.Items {
+			if strings.HasPrefix(pod.GetName(), cfg.podPrefix) {
+				podFound = true
+				break
+			}
+		}
+		if !podFound {
+			t.Fatalf("No pods with prefix %q found running in namespace %q", cfg.podPrefix, cfg.namespace)
 		}
 	}
-	return false
 }
 
 func createTestNamespace(client *kubernetes.Clientset, namespaceName string) (*v1.Namespace, error) {
@@ -278,6 +307,258 @@ func cleanupConfigMapCABundleInjectionTestObjects(client *kubernetes.Clientset,
 	// it should probably fail the test if the namespace gets stuck
 }
 
+type triggerRotationFunc func(*testing.T, *kubernetes.Clientset, *rest.Config)
+
+func checkCARotation(t *testing.T, client *kubernetes.Clientset, config *rest.Config, triggerRotation triggerRotationFunc) {
+	ns, err := createTestNamespace(client, "test-"+randSeq(5))
+	if err != nil {
+		t.Fatalf("could not create test namespace: %v", err)
+	}
+
+	// Prompt the creation of a service cert secret
+	testServiceName := "test-service-" + randSeq(5)
+	testSecretName := "test-secret-" + randSeq(5)
+	defer cleanupServiceSignerTestObjects(client, testSecretName, testServiceName, ns.Name)
+
+	err = createServingCertAnnotatedService(client, testSecretName, testServiceName, ns.Name)
+	if err != nil {
+		t.Fatalf("error creating annotated service: %v", err)
+	}
+
+	// Prompt the injection of the ca bundle into a configmap
+	testConfigMapName := "test-configmap-" + randSeq(5)
+	defer cleanupConfigMapCABundleInjectionTestObjects(client, testConfigMapName, ns.Name)
+
+	err = createAnnotatedCABundleInjectionConfigMap(client, testConfigMapName, ns.Name)
+	if err != nil {
+		t.Fatalf("error creating annotated configmap: %v", err)
+	}
+
+	// Retrieve the pre-rotation service cert
+	oldCertPEM, oldKeyPEM, err := pollForUpdatedServingCert(t, client, ns.Name, testSecretName, pollTimeout, nil, nil)
+	if err != nil {
+		t.Fatalf("error retrieving service cert: %v", err)
+	}
+
+	// Retrieve the pre-rotation ca bundle
+	oldBundlePEM, err := pollForUpdatedCABundle(t, client, ns.Name, testConfigMapName, pollTimeout, nil)
+	if err != nil {
+		t.Fatalf("error retrieving ca bundle: %v", err)
+	}
+
+	// Prompt CA rotation
+	triggerRotation(t, client, config)
+
+	// Retrieve the post-rotation service cert
+	newCertPEM, newKeyPEM, err := pollForUpdatedServingCert(t, client, ns.Name, testSecretName, rotationTimeout, oldCertPEM, oldKeyPEM)
+	if err != nil {
+		t.Fatalf("error retrieving service cert: %v", err)
+	}
+
+	// Retrieve the post-rotation ca bundle
+	newBundlePEM, err := pollForUpdatedCABundle(t, client, ns.Name, testConfigMapName, rotationTimeout, oldBundlePEM)
+	if err != nil {
+		t.Fatalf("error retrieving ca bundle: %v", err)
+	}
+
+	// Determine the dns name valid for the serving cert
+	certs, err := util.PemToCerts(newCertPEM)
+	if err != nil {
+		t.Fatalf("error decoding pem to certs: %v", err)
+	}
+	dnsName := certs[0].Subject.CommonName
+
+	util.CheckRotation(t, dnsName, oldCertPEM, oldKeyPEM, oldBundlePEM, newCertPEM, newKeyPEM, newBundlePEM)
+}
+
+// triggerTimeBasedRotation replaces the current CA cert with one that
+// is not valid for the minimum required duration and waits for the CA
+// to be rotated.
+func triggerTimeBasedRotation(t *testing.T, client *kubernetes.Clientset, config *rest.Config) {
+	// A rotation-prompting CA cert needs to be a renewed instance
+	// (i.e. share the same public and private keys) of the current
+	// cert to ensure that trust will be maintained for unrefreshed
+	// clients and servers.
+
+	// Retrieve current CA
+	secret, err := client.CoreV1().Secrets(serviceCAControllerNamespace).Get(signingKeySecretName, metav1.GetOptions{})
+	if err != nil {
+		t.Fatalf("error retrieving signing key secret: %v", err)
+	}
+	currentCACerts, err := util.PemToCerts(secret.Data[v1.TLSCertKey])
+	if err != nil {
+		t.Fatalf("error unmarshaling %q: %v", v1.TLSCertKey, err)
+	}
+	currentCAKey, err := util.PemToKey(secret.Data[v1.TLSPrivateKeyKey])
+	if err != nil {
+		t.Fatalf("error unmarshalling %q: %v", v1.TLSPrivateKeyKey, err)
+	}
+	currentCAConfig := &crypto.TLSCertificateConfig{
+		Certs: currentCACerts,
+		Key:   currentCAKey,
+	}
+
+	// Trigger rotation by renewing the current ca with an expiry that
+	// is sooner than the minimum required duration.
+	renewedExpiry := time.Now().Add(1 * time.Hour)
+	renewedCAConfig, err := util.RenewSelfSignedCertificate(currentCAConfig, renewedExpiry)
+	if err != nil {
+		t.Fatalf("error renewing ca to half-expired form: %v", err)
+	}
+	renewedCACertPEM, renewedCAKeyPEM, err := renewedCAConfig.GetPEMBytes()
+	if err != nil {
+		t.Fatalf("error encoding renewed ca to pem: %v", err)
+	}
+
+	// Write the renewed CA
+	secret = &v1.Secret{
+		Type: v1.SecretTypeTLS,
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      signingKeySecretName,
+			Namespace: serviceCAControllerNamespace,
+		},
+		Data: map[string][]byte{
+			v1.TLSCertKey:       renewedCACertPEM,
+			v1.TLSPrivateKeyKey: renewedCAKeyPEM,
+		},
+	}
+	_, _, err = resourceapply.ApplySecret(client.CoreV1(), events.NewInMemoryRecorder("test"), secret)
+	if err != nil {
+		t.Fatalf("error updating secret with test CA: %v", err)
+	}
+
+	pollForCARotation(t, client, renewedCACertPEM, renewedCAKeyPEM)
+}
+
+// triggerForcedRotation forces the rotation of the current CA via the
+// operator config.
+func triggerForcedRotation(t *testing.T, client *kubernetes.Clientset, config *rest.Config) {
+	// Retrieve the cert and key PEM of the current CA to be able to
+	// detect when rotation has completed.
+	secret, err := client.CoreV1().Secrets(serviceCAControllerNamespace).Get(signingKeySecretName, metav1.GetOptions{})
+	if err != nil {
+		t.Fatalf("error retrieving signing key secret: %v", err)
+	}
+	caCertPEM := secret.Data[v1.TLSCertKey]
+	caKeyPEM := secret.Data[v1.TLSPrivateKeyKey]
+
+	// Trigger a forced rotation by updating the operator config
+	operatorClient, err := operatorv1client.NewForConfig(config)
+	if err != nil {
+		t.Fatalf("error creating operator client: %v", err)
+	}
+	operatorConfig, err := operatorClient.OperatorV1().ServiceCAs().Get(api.OperatorConfigInstanceName, metav1.GetOptions{})
+	if err != nil {
+		t.Fatalf("error retrieving operator config: %v", err)
+	}
+	rawUnsupportedServiceCAConfig, err := operator.RawUnsupportedServiceCAConfig("42")
+	if err != nil {
+		t.Fatalf("failed to create raw unsupported config overrides: %v", err)
+	}
+	operatorConfig.Spec.UnsupportedConfigOverrides.Raw = rawUnsupportedServiceCAConfig
+	_, err = operatorClient.OperatorV1().ServiceCAs().Update(operatorConfig)
+	if err != nil {
+		t.Fatalf("error updating operator config: %v", err)
+	}
+
+	pollForCARotation(t, client, caCertPEM, caKeyPEM)
+}
+
+// pollForCARotation polls for the signing secret to be changed in
+// response to CA rotation.
+func pollForCARotation(t *testing.T, client *kubernetes.Clientset, caCertPEM, caKeyPEM []byte) {
+	_, err := pollForUpdatedSecret(t, client, serviceCAControllerNamespace, signingKeySecretName, pollTimeout, map[string][]byte{
+		v1.TLSCertKey:           caCertPEM,
+		v1.TLSPrivateKeyKey:     caKeyPEM,
+		api.BundleDataKey:       nil,
+		api.IntermediateDataKey: nil,
+	})
+	if err != nil {
+		t.Fatalf("error waiting for CA rotation: %v", err)
+	}
+}
+
+// pollForUpdatedServingCert returns the cert and key PEM if it changes from
+// that provided before the polling timeout.
+func pollForUpdatedServingCert(t *testing.T, client *kubernetes.Clientset, namespace, name string, timeout time.Duration, oldCertValue, oldKeyValue []byte) ([]byte, []byte, error) {
+	secret, err := pollForUpdatedSecret(t, client, namespace, name, timeout, map[string][]byte{
+		v1.TLSCertKey:       oldCertValue,
+		v1.TLSPrivateKeyKey: oldKeyValue,
+	})
+	if err != nil {
+		return nil, nil, err
+	}
+	return secret.Data[v1.TLSCertKey], secret.Data[v1.TLSPrivateKeyKey], nil
+}
+
+// pollForUpdatedSecret returns the given secret if its data changes from
+// that provided before the polling timeout.
+func pollForUpdatedSecret(t *testing.T, client *kubernetes.Clientset, namespace, name string, timeout time.Duration, oldData map[string][]byte) (*v1.Secret, error) {
+	resourceID := fmt.Sprintf("Secret \"%s/%s\"", namespace, name)
+	obj, err := pollForResource(t, resourceID, timeout, func() (kruntime.Object, error) {
+		secret, err := client.CoreV1().Secrets(namespace).Get(name, metav1.GetOptions{})
+		if err != nil {
+			return nil, err
+		}
+		err = util.CheckData(oldData, secret.Data)
+		if err != nil {
+			return nil, err
+		}
+		return secret, nil
+	})
+	if err != nil {
+		return nil, err
+	}
+	return obj.(*v1.Secret), nil
+}
+
+// pollForUpdatedCABundle returns the given configmap if its data changes from
+// that provided before the polling timeout.
+func pollForUpdatedCABundle(t *testing.T, client *kubernetes.Clientset, namespace, name string, timeout time.Duration, oldValue []byte) ([]byte, error) {
+	resourceID := fmt.Sprintf("ConfigMap \"%s/%s\"", namespace, name)
+	expectedDataSize := 1
+	obj, err := pollForResource(t, resourceID, timeout, func() (kruntime.Object, error) {
+		configMap, err := client.CoreV1().ConfigMaps(namespace).Get(name, metav1.GetOptions{})
+		if err != nil {
+			return nil, err
+		}
+		if len(configMap.Data) != expectedDataSize {
+			return nil, fmt.Errorf("expected data size %d, got %d", expectedDataSize, len(configMap.Data))
+		}
+		value, ok := configMap.Data[api.InjectionDataKey]
+		if !ok {
+			return nil, fmt.Errorf("key %q is missing", api.InjectionDataKey)
+		}
+		if value == string(oldValue) {
+			return nil, fmt.Errorf("value for key %q has not changed", api.InjectionDataKey)
+		}
+		return configMap, nil
+	})
+	if err != nil {
+		return nil, err
+	}
+	configMap := obj.(*v1.ConfigMap)
+	return []byte(configMap.Data[api.InjectionDataKey]), nil
+}
+
+// pollForResource returns a kruntime.Object if the accessor returns without error before the timeout.
+func pollForResource(t *testing.T, resourceID string, timeout time.Duration, accessor func() (kruntime.Object, error)) (kruntime.Object, error) {
+	var obj kruntime.Object
+	err := wait.PollImmediate(pollInterval, timeout, func() (bool, error) {
+		o, err := accessor()
+		if err != nil && errors.IsNotFound(err) {
+			return false, nil
+		}
+		if err != nil {
+			t.Logf("an error occurred while polling for %s: %v", resourceID, err)
+			return false, nil
+		}
+		obj = o
+		return true, nil
+	})
+	return obj, err
+}
+
 func TestE2E(t *testing.T) {
 	// use /tmp/admin.conf (placed by ci-operator) or KUBECONFIG env
 	confPath := "/tmp/admin.conf"
@@ -301,18 +582,7 @@ func TestE2E(t *testing.T) {
 
 	// the service-serving-cert-operator and controllers should be running as a stock OpenShift component. our first test is to
 	// verify that all of the components are running.
-	if !hasPodWithPrefixName(adminClient, serviceCAOperatorPodPrefix, serviceCAOperatorNamespace) {
-		t.Fatalf("%s not running in %s namespace", serviceCAOperatorPodPrefix, serviceCAOperatorNamespace)
-	}
-	if !hasPodWithPrefixName(adminClient, apiInjectorPodPrefix, serviceCAControllerNamespace) {
-		t.Fatalf("%s not running in %s namespace", apiInjectorPodPrefix, serviceCAControllerNamespace)
-	}
-	if !hasPodWithPrefixName(adminClient, configMapInjectorPodPrefix, serviceCAControllerNamespace) {
-		t.Fatalf("%s not running in %s namespace", configMapInjectorPodPrefix, serviceCAControllerNamespace)
-	}
-	if !hasPodWithPrefixName(adminClient, caControllerPodPrefix, serviceCAControllerNamespace) {
-		t.Fatalf("%s not running in %s namespace", caControllerPodPrefix, serviceCAControllerNamespace)
-	}
+	checkComponents(t, adminClient)
 
 	// test the main feature. annotate service -> created secret
 	t.Run("serving-cert-annotation", func(t *testing.T) {
@@ -540,6 +810,22 @@ func TestE2E(t *testing.T) {
 		}
 	})
 
+	// This test triggers rotation by updating the CA to have an
+	// expiry that is less than the minimum required duration and then
+	// validates that both refreshed and unrefreshed clients and
+	// servers can continue to communicate in a trusted fashion.
+	t.Run("time-based-ca-rotation", func(t *testing.T) {
+		checkCARotation(t, adminClient, nil, triggerTimeBasedRotation)
+	})
+
+	// This test triggers rotation by updating the operator
+	// configuration to force rotation and then validates that both
+	// refreshed and unrefreshed clients and servers can continue to
+	// communicate in a trusted fashion.
+	t.Run("forced-ca-rotation", func(t *testing.T) {
+		checkCARotation(t, adminClient, adminConfig, triggerForcedRotation)
+	})
+
 	// TODO: additional tests
 	// - API service CA bundle injection
 }
